﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Testing;

namespace OpcLabs.UnitTestLib.Testers
{
    public class EqualityOperatorsTester : Tester
    {
        public EqualityOperatorsTester() 
        {
            TestCategories.Add("Equality");
        }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            int i = 0;
            foreach (KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator in InstanceCreators)
            {
                string instanceName = instanceCreator.Key;
                Debug.Assert(instanceName != null);

                AddTestMethod(CreateReflexivityTestMethod(typeUnderTest, instanceName));
                if (i < 1)
                    AddTestMethod(CreateWithNullTestMethod(typeUnderTest, instanceName));

                i++;
            }

            if (GetOption("UnequalTest"))
                AddTestMethod(CreateUnequalTestMethod(typeUnderTest));
        }

        [NotNull]
        private TestMethod CreateReflexivityTestMethod([NotNull] Type typeUnderTest, [NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "Reflexivity_{0}", instanceName), 
                () =>
                    {
                        MethodInfo equalityOperator = typeUnderTest.GetEqualityOperator();
                        Assert.IsNotNull(equalityOperator);

                        MethodInfo inequalityOperator = typeUnderTest.GetInequalityOperator();
                        Assert.IsNotNull(inequalityOperator);

                        object instance = CreateInstance(instanceName);
                        AssertUtilities.WriteLine(instance.GetHashCode());

                        {
                            object equalityResult = equalityOperator.Invoke(null, new[] {instance, instance});
                            Assert.IsNotNull(equalityResult);
                            Assert.IsInstanceOfType(equalityResult, typeof (bool));
                            Assert.IsTrue((bool) equalityResult);
                        }

                        {
                            object inequalityResult = inequalityOperator.Invoke(null, new[] {instance, instance});
                            Assert.IsNotNull(inequalityResult);
                            Assert.IsInstanceOfType(inequalityResult, typeof (bool));
                            Assert.IsFalse((bool) inequalityResult);
                        }
                    });
        }

        [NotNull]
        private TestMethod CreateUnequalTestMethod([NotNull] Type typeUnderTest)
        {
            return new DelegatedTestMethod(
                "Unequal",
                () =>
                {
                    MethodInfo equalityOperator = typeUnderTest.GetEqualityOperator();
                    Assert.IsNotNull(equalityOperator, "Equality operator not found.");

                    MethodInfo inequalityOperator = typeUnderTest.GetInequalityOperator();
                    Assert.IsNotNull(inequalityOperator, "Inequality operator not found.");

                    if (InstanceCreators.Count <= 1)
                        AssertUtilities.Inconclusive("Not enough instances available ({0}). At least two instances are needed.", InstanceCreators.Count);

                    foreach (KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator1 in InstanceCreators)
                    {
                        string instanceName1 = instanceCreator1.Key;
                        Debug.Assert(instanceName1 != null);

                        foreach (
                            KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator2 in InstanceCreators)
                        {
                            string instanceName2 = instanceCreator2.Key;
                            Debug.Assert(instanceName2 != null);

                            if (instanceName2 == instanceName1)
                                continue;

                            AssertUtilities.WriteLine($"{instanceName1}, {instanceName2}");

                            object instance1 = CreateInstance(instanceName1);
                            AssertUtilities.WriteLine(instance1.GetHashCode());

                            object instance2 = CreateInstance(instanceName2);
                            AssertUtilities.WriteLine(instance2.GetHashCode());

                            {
                                object equalityResult = equalityOperator.Invoke(null, new[] {instance1, instance2});
                                Assert.IsNotNull(equalityResult);
                                Assert.IsInstanceOfType(equalityResult, typeof (bool));
                                Assert.IsFalse((bool) equalityResult, "Equality operator returned true but it should have returned false.");
                            }

                            {
                                object inequalityResult = inequalityOperator.Invoke(null, new[] {instance1, instance2});
                                Assert.IsNotNull(inequalityResult);
                                Assert.IsInstanceOfType(inequalityResult, typeof (bool));
                                Assert.IsTrue((bool) inequalityResult, "Inequality operator returned false but it should have returned true.");
                            }
                        }
                    }
                });
        }

        [NotNull]
        private TestMethod CreateWithNullTestMethod([NotNull] Type typeUnderTest, [NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "WithNull_{0}", instanceName),
                () =>
                {
                    MethodInfo equalityOperator = typeUnderTest.GetEqualityOperator();
                    Assert.IsNotNull(equalityOperator);

                    MethodInfo inequalityOperator = typeUnderTest.GetInequalityOperator();
                    Assert.IsNotNull(inequalityOperator);

                    object instance = CreateInstance(instanceName);

                    {
                        object equalityResult1 = equalityOperator.Invoke(null, new[] {instance, null});
                        Assert.IsNotNull(equalityResult1);
                        Assert.IsInstanceOfType(equalityResult1, typeof (bool));
                        Assert.IsFalse((bool) equalityResult1);
                    }

                    {
                        object inequalityResult1 = inequalityOperator.Invoke(null, new[] {instance, null});
                        Assert.IsNotNull(inequalityResult1);
                        Assert.IsInstanceOfType(inequalityResult1, typeof (bool));
                        Assert.IsTrue((bool) inequalityResult1);
                    }

                    {
                        object equalityResult2 = equalityOperator.Invoke(null, new[] {null, instance});
                        Assert.IsNotNull(equalityResult2);
                        Assert.IsInstanceOfType(equalityResult2, typeof (bool));
                        Assert.IsFalse((bool) equalityResult2);
                    }

                    {
                        object inequalityResult2 = inequalityOperator.Invoke(null, new[] {null, instance});
                        Assert.IsNotNull(inequalityResult2);
                        Assert.IsInstanceOfType(inequalityResult2, typeof (bool));
                        Assert.IsTrue((bool) inequalityResult2);
                    }
                });
        }

    }
}
