﻿
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.BaseLib.Runtime.InteropServices;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.Engine;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    partial class EasyDAClientTestBase
    {
        #region Static Protected Methods

        static protected void AssertBaseException(
            [NotNull] Exception exception,
            int expectedErrorCode,
            [NotNull] string expectedNamespace,
            [NotNull] string expectedName)
        {
            Exception baseException = exception.GetBaseException();
            Assert.IsNotNull(baseException);
            AssertUtilities.WriteLine($"baseException: {baseException}");
            Assert.IsInstanceOfType(baseException, typeof(ExtendedCOMException));
            var extendedComException = (ExtendedCOMException)baseException;
            AssertUtilities.WriteLine($"extendedComException.QualifiedName: {extendedComException.QualifiedName}");
            AssertUtilities.WriteLine($"extendedComException.GetErrorId(): {extendedComException.GetErrorId()}");
            // ReSharper disable once CommentTypo
            Assert.AreEqual(expectedErrorCode, extendedComException.ErrorCode);
            Assert.AreEqual(expectedNamespace, extendedComException.QualifiedName.Namespace);
            Assert.AreEqual(expectedName, extendedComException.QualifiedName.Name);
        }

        #endregion



        #region Protected Methods

        protected void ReadItemCatalogTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemCatalogEntry itemCatalogEntry)
        {
            VarType dataType = itemCatalogEntry.DataType;
            Type expectedType = null;
            if (dataType != VarTypes.Empty)
                expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));

            ReadItemTest_Basic(serverDescriptor, itemCatalogEntry.Element, expectedType);
        }

        protected void ReadItemTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            [CanBeNull] Type expectedType)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            EasyDAClient client = CreateClient();

            TestSection.Act();
            DAVtq actualVtq;
            try
            {
                actualVtq = client.ReadItem(serverDescriptor, itemDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            TestSection.Assert();

            Assert.IsNotNull(actualVtq);
            AssertUtilities.WriteLine($"actualVtq: {actualVtq}");

            if (expectedType != null)
            {
                object actualValue = actualVtq.Value;
                Assert.IsNotNull(actualValue);
                Assert.AreEqual(expectedType, actualValue.GetType());
            }

            Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
            DateTime dateTime0 = DateTime.FromFileTimeUtc(0);
            AssertUtilities.WriteLine($"dateTime0: {dateTime0}");
            Assert.IsTrue(actualVtq.Timestamp >= dateTime0, "actualVtq.Timestamp >= dateTime0");
            DateTime utcNow = DateTime.UtcNow;
            AssertUtilities.WriteLine($"UtcNow: {utcNow}");
            TimeSpan timeTolerance = TimeSpan.FromSeconds(60);
            Assert.IsTrue(actualVtq.Timestamp <= utcNow + timeTolerance, "actualVtq.Timestamp <= utcNow + timeTolerance");
        }

        protected void ReadItemTest_NonReadable(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            EasyDAClient client = CreateClient();

            TestSection.Act();
            try
            {
                try
                {
                    client.ReadItem(serverDescriptor, itemDescriptor);
                }
                finally
                {
                    TestSection.Assert();
                }
            }
            catch (OpcException opcException)
            {
                AssertUtilities.WriteLine(opcException.Message ?? String.Empty);
                Exception innerException = opcException.InnerException;
                Assert.IsNotNull(innerException);
                AssertUtilities.WriteLine(innerException.GetType());
                string innerExceptionMessage = innerException.Message;
                AssertUtilities.WriteLine(innerExceptionMessage ?? String.Empty);
                Assert.IsInstanceOfType(innerException, typeof(NetApiResultException));
                // ReSharper disable once StringLiteralTypo
                StringAssert.Contains(innerExceptionMessage, "\"E_WRITEONLY\"");
                return;
            }
            Assert.Fail();
        }

        protected void ReadItemTest_UnknownItemId(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            EasyDAClient client = CreateClient();

            TestSection.Act();
            try
            {
                try
                {
                    client.ReadItem(serverDescriptor, itemDescriptor);
                }
                finally
                {
                    TestSection.Assert();
                }
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                Exception innerException = opcException.InnerException;
                Assert.IsNotNull(innerException);
                AssertBaseException(innerException, 
                    -1073479673, // OPC_E_UNKNOWNITEMID
                    "http://opcfoundation.org/DataAccess/",
                    "E_UNKNOWN_ITEM_NAME");
                return;
            }
            Assert.Fail();
        }

        protected void ReadMultipleItemsCatalogTest(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] IEnumerable<DAItemCatalogEntry> itemCatalogEntries)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);

            DAItemCatalogEntry[] itemCatalogEntryArray = itemCatalogEntries.ToArray();
            var argumentsList = new List<DAReadItemArguments>();
            foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntryArray)
            {
                Debug.Assert(itemCatalogEntry != null);

                var arguments = new DAReadItemArguments(serverDescriptor, itemCatalogEntry.Element);
                argumentsList.Add(arguments);
            }

            DAReadItemArguments[] argumentsArray = argumentsList.ToArray();
            int count = argumentsArray.Length;

            EasyDAClient client = CreateClient();
            var stopwatch = new Stopwatch();

            TestSection.Act();
            stopwatch.Start();
            DAVtqResult[] resultArray = client.ReadMultipleItems(argumentsArray);
            stopwatch.Stop();

            TestSection.Assert();
            AssertUtilities.WriteLine(stopwatch.Elapsed);

            Assert.IsNotNull(resultArray);
            Assert.AreEqual(count, resultArray.Length);
            for (int i = 0; i < count; i++)
            {
                DAItemCatalogEntry itemCatalogEntry = itemCatalogEntryArray[i];
                Debug.Assert(itemCatalogEntry != null);

                AssertUtilities.WriteLine("");
                AssertUtilities.WriteLine(i);
                AssertUtilities.WriteLine(itemCatalogEntry.Element);

                DAVtqResult result = resultArray[i];
                Assert.IsNotNull(result);
                AssertUtilities.WriteLine(result);

                if (!client.SupportsTechnology(serverDescriptor.Technology))
                {
                    Assert.IsFalse(result.Succeeded);
                    Assert.IsInstanceOfType(result.Exception, typeof(NotSupportedException));
                    continue;
                }

                if (itemCatalogEntry.AccessRights.AllowRead)
                {
                    Assert.IsNull(result.Exception);
                    Assert.AreEqual(null, result.ErrorId);

                    DAVtq actualVtq = result.Vtq;
                    Assert.IsNotNull(actualVtq);

                    Type expectedType = null;

                    VarType dataType = itemCatalogEntry.DataType;
                    if (dataType != VarTypes.Empty)
                        expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));
                    if (expectedType != null)
                    {
                        object actualValue = actualVtq.Value;
                        Assert.IsNotNull(actualValue);
                        Assert.AreEqual(expectedType, actualValue.GetType());
                    }

                    Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
                    Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
                    Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
                    Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
                    DateTime utcNow = DateTime.UtcNow;
                    AssertUtilities.WriteLine($"UtcNow: {utcNow}");
                    TimeSpan timeTolerance = TimeSpan.FromSeconds(60);
                    Assert.IsTrue(actualVtq.Timestamp <= utcNow + timeTolerance, "actualVtq.Timestamp <= utcNow + timeTolerance");
                }
                else
                {
                    Assert.IsNotNull(result.Exception);
                    // IMPROVE: test more about the exception

                    Assert.IsNull(result.Vtq);
                }
            }
        }

        #endregion
    }
}
